"""
================================================================================
Customize PyEdit popup windows created by PyMailGUI (not View windows).


SUMMARY:

This file's settings (Python-coded assignments) customize PyEdit popup
windows produced by PyMailGUI for contexts such as text-part, HTML-part,
and raw-message views.  This file does not apply to the basic aspects of
PyMailGUI's View and Write windows, configured in folder ../MailConfigs.

Note that font, color, and size can be changed once the GUI window is open
(via menu options and resizes), so many of these are initial values only,
but are applied at the start of every session for convenience.  Color and
font list settings provide access to your favorite options in every session.
See PP4E/Gui/TextEditor/textConfig.py for more on the available settings.

This module (not its package) is assumed to be on the path for its settings:
it _replaces_ PyEdit's own config file (at PP4E/Gui/TextEditor/textConfig.py)
when PyEdit is used by PyMailGUI, though for simplicity PyEdit's Unicode
settings always come from its own package's textConfig.py, not this file.
All other unset names here use in-program presets, not PyEdit's own textConfig.

UPDATE: this file now runs an initial "from *" to load all settings from 
PyEdit's own textConfig.py in the PP4E package.  Hence, assignments here 
now _do_ override the base set in PyEdit's config file.  This was done to 
avoid code redundancy and its inevitable maintenance overheads.  Names
unset both here _and_ in PyEdit's version still invoke program defaults.


WHAT SETTINGS HERE APPLY TO:

PyMailGUI uses PyEdit in two different roles -- for popup text windows, and
as an embedded component inside View/Write windows.  These two PyEdit roles
are configured by different setting sources in PyMailGUI:

1) Popups: font, color, and size settings in this file ("textConfig.py") apply
only to text displayed by PyMailGUI's standalone text popups.  These popups are
created for raw text on double clicks, plain and HTML text parts selected, etc.

2) Embedded: font, color, and size settings in "mailconfig.py" apply only to
text displayed in PyMailGUI's own View/Write windows.  These settings are also
customized by per-account files in the top-level launcher's MailConfigs folder. 

Apart from initial fonts, colors, and sizes, settings here apply to *both*
PyEdit contexts in PyMailGUI - popup and embedded - because this file replaces
PyEdit's own version.  For example, color and font lists set here will apply
to both PyEdit roles in PyMailGUI; if omitted both here and in the imported
PyEdit version of this file, they will default to PyEdit's presets.


ASSORTED NOTES:

1) Caution: edit with care, as PyEdit currently does not error-check any of
these settings (this is a to-do); errors are silently ignored or hang the GUI.
Be sure to run or import this file to check for syntax errors (hint: you can
use PyEdit's own Tools=>Run Code for this purpose).

2) Caveat: items here are per install, and not configurable per account in the
top-level MailConfigs folder (PyEdit imports them from PyMailGUI's '.' here).
Use MailConfigs settings to distinguish accounts' GUIs if desired.


UNICODE SETTINGS AND POPUPS:

As shipped, PyEdit uses popups to always ask for Unicode text encodings on
both opens and saves, instead of assuming the platform's default.  This is
flexible and avoids inaccurate guesses, but many users can safely accept the
platform default displayed in the popup, especially for files they create.

To change PyEdit's Unicode policies to skip the popups altogether, edit the
Unicode settings near the end of the PyEdit folder's own config file, at:

    PP4E/Gui/TextEditor/textConfig.py

For example, "opensAskUser" and "savesAskUser" control the file-open and
file-save popups, respectively; set these to False to silence the popups.
These settings always come from PyEdit's file, and cannot be given here.
Users who edit mixed Unicode-type files may prefer to retain the popups.
================================================================================
"""

# for checking platform, version, etc.
import sys   

# get default settings from PyEdit's version of this file
from PP4E.Gui.TextEditor.textConfig import *

# for platform-specific choices
RunningOnMac     = sys.platform.startswith('darwin')       # all OS (X)
RunningOnWindows = sys.platform.startswith('win')          # all Windows
RunningOnLinux   = sys.platform.startswith('linux')        # all Linux


#===============================================================================
# initial background, foreground (and cursor) colors
#===============================================================================

bg = 'beige'     # initial window color: absent=white; colorname or RGB hexstr
fg = 'black'     # initial text color:   absent=black;  e.g., 'beige', '#690f96'


#===============================================================================
# initial text-area font
#===============================================================================

if RunningOnMac:                       # [4.0] specialize on Mac, else too small
    font = ('menlo', 13, 'normal')     # also mono on Mac: 'monaco', courier/14

elif RunningOnWindows:
    font = ('courier', 12, 'bold')     # or consolas; e.g., style: 'bold italic'

elif RunningOnLinux:
    font = ('courier', 12, 'bold')     # or incosolata; same font as Windows?


#===============================================================================
# auto-save dir, or None to turn off saves [4.0]
# in PyMailGUI, applies to both View/Write and text-popup windows' text
#===============================================================================

autoSaveToFolder = '__pyedit-autosaves__'    # in '.', or a pathname provided


#===============================================================================
# pick next color in colorlist for each now PyEdit popup and component? [4.0]
# preset False: MailConfigs colors are per-account (cycled colors here are not);
#===============================================================================

colorCycling = False                         # True to enable cycling



#===============================================================================
# ETC - the following all have either defaults in PyEdit's own version of
# this file or in-program presets used if not assigned here.  To customize
# here for PyMailGUI, uncomment (remove leading "# ") and edit.  This file's
# settings replace those in PyEdit's own textConfig (except for Unicode names),
# so names not set here use either in-program or PyEdit's textConfig defaults.
# See PP4E\Gui\TextEditor\textConfig.py for more details, and Unicode policies.   
#
# height = 20                                       # win size, default=24 lines
# width  = 84                                       # win size, default=80 chars
#
# toolbarFont = None                                # for button row at bottom
# toolbarFixedLayout = False                        # spacers do not expand?
#
# runcodefont = None                                # font: Run Code output text
# runcodebg = runcodefg = None                      # colors: Run Code output
#
# caseinsens = True                                 # Find searches ignore case?
#
# colorlist = [dict(fg='black',  bg='white'),       # Color List presets [4.0]
#              dict(fg='green2', bg='black'),       # color name or #RRGGBB hex
#              dict(fg='white',  bg='#173166')      # as many as you wish
#              ]                                    # picks next each time
#
# fontlist = [('courier', 11, 'normal'),            # Font List presets [4.0]
#             ('arial',   12, 'bold italic'),       # (family, size, 'style...')
#             ('times',   18, 'bold')               # as many as you wish 
#             ]                                     # picks next each time
#
# grepSpawnMode = 'multiprocessing'                 # starts new process? [4.0]
#
# autoSaveMinsTillRun  = 5                          # minutes between auto-saves
# autoSaveDaysRetained = 7                          # days until files deleted
#
# RunCode_PYTHONEXECUTABLE = None                   # pathname to alt python 
# RunCode_PYTHONPATH_APPENDS = []                   # pathname strings to add
# RunCode_PYTHONPATH_PREPENDS = []                  # pathname strings to add
#===============================================================================
